#include<GL/glut.h>
#include<math.h>
#include"camera3d.h"

Camera3D::Camera3D(void)
{
  interactiva = !false;

    // Parametros del modo interactivo
  angleA   = 0.0;              // Giro alrededor de las Y (radianes)
  angleB   = 0.0;              // Inclinacion del cabezon (radianes)
  velocity = 3.0;              // Velocidad
  position.set(0, 40, 400);     // Posicion del observador
  incr.set(0, 0, velocity);
  prelookat.set(0, 0, -1);
  

      // Camara automatica
  rx = M_PI*220/100;           // Angulo de giro (radianes)
  dist = 400;                  // Distancia de la camara al punto que miramos
  lookat = Vector3D(0, 0, 0);  // centro alrededor del cual giramos

    // Parametros de la camara
  angle = 60.0;                // Angulo de obertura de la camara
  znear = 0.1;                 // Plano de corte anterior
  zfar = 8000;                 // Plano de corte anterior
  aspect = 1;                  // Aspect-Ratio
  zoom = 1;                    // Factor de zoom
}

void Camera3D::setZnear(float zn) { znear = zn; }
void Camera3D::setZfar(float zf)  { zfar = zf; }

void Camera3D::avanzar(bool palante = true)
{
  position = (palante) ? position - incr : position + incr;
  prelookat = Vector3D(sin(angleA)*cos(angleB), sin(angleB), -cos(angleA)*cos(angleB)) + position;
}

void Camera3D::pan(bool paloslaos, bool inc)
{
  register float v = (inc) ? velocity : -velocity;
  if(paloslaos)
    position = Vector3D(v*cos(angleA), 0, v*sin(angleA)) + position;
  else
    position = Vector3D(0, -v*cos(angleB), 0) + position;
  prelookat = Vector3D(sin(angleA)*cos(angleB), sin(angleB), -cos(angleA)*cos(angleB)) + position;
}

void Camera3D::setVel(float vel)
{
  velocity = vel;
  incr = Vector3D(-velocity*sin(angleA)*cos(angleB), -velocity*sin(angleB), velocity*cos(angleA)*cos(angleB));
}

void Camera3D::girar(float izqder = 0.0, float arrbaj = 0.0)
{
  angleA += izqder;
  if(angleA <= -(2.0*M_PI)) angleA += 2.0*M_PI;
  if(angleA >=  (2.0*M_PI)) angleA -= 2.0*M_PI;
  angleB += arrbaj;
  if(angleB <= -(M_PI/2.0)) angleB = -M_PI/2.0 * .99;
  if(angleB >=  (M_PI/2.0)) angleB =  M_PI/2.0 * .99;
  prelookat = Vector3D(sin(angleA)*cos(angleB), sin(angleB), -cos(angleA)*cos(angleB)) + position;
  incr = Vector3D(-velocity*sin(angleA)*cos(angleB), -velocity*sin(angleB), velocity*cos(angleA)*cos(angleB));
}

void Camera3D::setPosition(Vector3D pos)
{
  position = pos;
}

void Camera3D::SetupProjectionOpenGL(void)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(angle * zoom, aspect, znear, zfar);
  glMatrixMode(GL_MODELVIEW);
}

void Camera3D::SetupCameraOpenGL(void)
{
  glLoadIdentity();
  if(!interactiva)
  {
    gluLookAt(dist * cos(rx), dist * (cos(rx/4) + 1.0), dist * sin(rx),  // Posicion (x,y,z) del observador
              lookat.X(), lookat.Y(), lookat.Z(),                        // Posicion del centro de los objetos a mostrar
              0.0, 1.0, 0.0);                                            // VIEW - UP - VECTOR (VUP)
  } else {
    gluLookAt(position.X(),  position.Y(),  position.Z(),  // Posicion (x,y,z) del observador
              prelookat.X(), prelookat.Y(), prelookat.Z(), // Punto al que miramos
              0.0, 1.0, 0.0);                              // VIEW - UP - VECTOR (VUP)
  }
}

void Camera3D::save(FILE *f)
{
  fwrite(&interactiva, sizeof(interactiva), 1, f);
  fwrite(&angleA,      sizeof(angleA),      1, f);
  fwrite(&angleB,      sizeof(angleB),      1, f);
  fwrite(&velocity,    sizeof(velocity),    1, f);
  position.save(f);
  incr.save(f);
  prelookat.save(f); 
  fwrite(&rx,          sizeof(rx),          1, f);
  fwrite(&dist,        sizeof(dist),        1, f);
  lookat.save(f);
  fwrite(&angle,       sizeof(angle),       1, f);
  fwrite(&znear,       sizeof(znear),       1, f);
  fwrite(&zfar,        sizeof(zfar),        1, f);
  fwrite(&aspect,      sizeof(aspect),      1, f);
  fwrite(&zoom,        sizeof(zoom),        1, f);
}

void Camera3D::load(FILE *f)
{
  fread(&interactiva, sizeof(interactiva), 1, f);
  fread(&angleA,      sizeof(angleA),      1, f);
  fread(&angleB,      sizeof(angleB),      1, f);
  fread(&velocity,    sizeof(velocity),    1, f);
  position.load(f);
  incr.load(f);
  prelookat.load(f); 
  fread(&rx,          sizeof(rx),          1, f);
  fread(&dist,        sizeof(dist),        1, f);
  lookat.load(f);
  fread(&angle,       sizeof(angle),       1, f);
  fread(&znear,       sizeof(znear),       1, f);
  fread(&zfar,        sizeof(zfar),        1, f);
  fread(&aspect,      sizeof(aspect),      1, f);
  fread(&zoom,        sizeof(zoom),        1, f);
}
