/****************************************************************************
 * Fichero: lab4.cpp
 *
 * Autor: Gerardo Garca Pea
 * Asignatura: GC (2002 - Cuadrimestre Primavera)
 *
 ****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <GL/glut.h>
#include "main.h"
#include "model3d.h"
#include "libmodel.h"
#include "loadvrml.h"
#include "util.h"

//////////////////////////////////////////////////////////////////////////////
// CONFIGURACION
//
//
//   PIECE_SEPARATION   - Separacin de las piezas / tamao de los cuadros
//   ANCHURA_BORDE_TABLERO
//   ALTURA_BORDE_TABLERO
//   FACTOR_HUNDIMIENTO - Parte del tablero que sobresale
//   SIMPLIFIED_SCENE   - Una versin de la escena MUY simple
//   TIMER_MSECS        - Cada cuantos milisecs se calcula el frame-rate
//   SUAVIDAD_GIRO      - Divisor que se aplica a los movimientos del raton
//   SUAVIDAD_PAN       - Idem
//   KEYB_VEL           - Velocidad de movimientos con el teclado
//   ZIP_CACHE          - utiliza gzip/gunzip para comprimir/descomprimir
//                        la escena
//
//////////////////////////////////////////////////////////////////////////////

  // Propiedades del ajedrez (caracteristicas)
#define PIECE_SEPARATION      55.0
#define ANCHURA_BORDE_TABLERO 30.0
#define ALTURA_BORDE_TABLERO  16.0
#define FACTOR_HUNDIMIENTO    (1.0 / 10.0)
//#define SIMPLIFIED_SCENE
#define ZIP_CACHE

  // Frame-Rate calculator (propiedades)
#define TIMER_MSECS 2000

  // Navegacin por la escena
#define SUAVIDAD_GIRO	500.0
#define SUAVIDAD_PAN    5.0
#define KEYB_VEL        10.0

  // Nombres de ficheros
#ifdef ZIP_CACHE
#define SCENE_O3D       "scene.o3d.gz"
#define SHADOWVOLS_O3D  "shadowvols.o3d.gz"
#else
#define SCENE_O3D       "scene.o3d"
#define SHADOWVOLS_O3D  "shadowvols.o3d"
#endif

//////////////////////////////////////////////////////////////////////////////
// MATERIALES DE LA ESCENA
//
//////////////////////////////////////////////////////////////////////////////

Material3D tblack(Color3D(0.100000, 0.100000, 0.100000),
                  Color3D(0.400000, 0.400000, 0.400000),
                  Color3D(0.700000, 0.700000, 0.700000),
                  true, 40.00000, 0.8, 1);
Material3D twhite(Color3D(0.600000, 0.600000, 0.500000),
                  Color3D(0.900000, 0.900000, 0.700000),
                  Color3D(0.800000, 0.800000, 0.600000),
                  true, 40.00000, 0.8, 1);


Material3D pblack(Color3D(0.200000, 0.200000, 0.200000),
                  Color3D(0.500000, 0.500000, 0.500000),
                  Color3D(1.000000, 1.000000, 1.000000),
                  false, 128.000000, 1.0);
Material3D pwhite(Color3D(0.500000, 0.500000, 0.400000),
                  Color3D(0.700000, 0.700000, 0.600000),
                  Color3D(0.800000, 0.800000, 0.600000),
                  false, 128.000000, 1.0);

Material3D madeira(Color3D(168.0 / (255.0 * 1.5), 121.0 / (255.0 * 1.5), 28.0 / (255.0 * 1.5)),
                   Color3D(168.0 / (255.0 * 1.8), 121.0 / (255.0 * 1.8), 28.0 / (255.0 * 1.6)),
                   Color3D(0, 0, 0),
                   false, 0.00000, 1.0);

//////////////////////////////////////////////////////////////////////////////
// VARIABLES GLOBALES
//
//  last_x, last_y   - Ultimo sitio donde se observo la posicion del raton
//  SUMRX_INC, sumrx - Incremento a aplicar en la camara automatica
//  bishop...        - modelos de la escena
//  frames           - # de frames contados con el timer
//  modelos[]        - Puntero a nuestras fichas de ajedrez
//  Tablero          - Tablero de ajedrez
//////////////////////////////////////////////////////////////////////////////

int last_x;
int last_y;
int doing_pan;
float SUMRX_INC = 0.04;
float sumrx = 0.0;
Model3D bishop, knight, pawn, queen, rook, king, board, // Las diferentes piezas del ajedrez y el tablero
        scene, shadowvols;                              // La escena y sus sombras

int frames = 0;
Model3D *modelos[6] = { &pawn, &knight, &bishop, &rook, &queen, &king };

#ifdef SIMPLIFIED_SCENE
int Tablero[8*8+1] = {
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  5,  1,  0,  0,  0,
         0,  0,  0, 11, 15,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0
};
#else
// NOTA: Es el mate del pastor. La jugada mas famosa del ajedrez.
int Tablero[8*8+1] = {
         4,  0,  3,  5,  6,  3,  2,  4,
         1,  1,  1,  0,  0,  1,  1,  1,
         0,  0,  2,  1,  0,  0,  0,  0,
         0,  0,  0,  0,  1,  0,  0,  0,
         0,  0, 13,  0, 11,  0,  0,  0,
         0,  0,  0,  0,  0, 15,  0,  0,
        11, 11, 11, 11,  0, 11, 11, 11,
        14, 12, 13,  0, 16,  0, 12, 14,
         0
};
#endif

void timerCallback(int value)
{
  cout << "Frame-rate: " << ((float) frames) / (((float) TIMER_MSECS) / 1000.0) << " frames/sec     \r";
  flush(cout);
  frames = 0;
  glutTimerFunc(TIMER_MSECS, timerCallback, value);
}

void CompruebaError(char *s)
{
  register GLenum error;
  
  if((error = glGetError()) != GL_NO_ERROR)
  {
    register char *s2 = (char *) ((s == NULL) ? "" : s);
    printf("Error:OpenGL:%s: Error [%d] detectado!\nError:OpenGL:%s: glutErrorString() = '%s'\n", s2, error, s2, gluErrorString(error));
  }
}

void callbackRender(void)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  scene.toOpenGL();

  glutSwapBuffers();
  frames++;

  CompruebaError("callbackRender()");
}

void callbackResize(int width, int height)
{
    /* Redefinimos el viewport */
  glViewport(0, 0, width, height);
    /* Recalculamos la matriz de proyeccion */
  scene.cam.aspect = ((float) width) / ((float) height);

  scene.cam.SetupProjectionOpenGL();
}

void callbackMove(void)
{
  if(sumrx != 0 && !scene.cam.interactiva)
  {
    scene.cam.rx += sumrx;
    glutPostRedisplay();
  }
}

void ComandosInteractivos(int key)
{
  scene.cam.setVel(KEYB_VEL);
  switch(key)
  {
    case '8':
    case 'W':
    case 'w': scene.cam.avanzar();         glutPostRedisplay(); break;
    case '5':
    case 'S':
    case 's': scene.cam.avanzar(false);    glutPostRedisplay(); break;
    case '4': scene.cam.girar(-0.1);       glutPostRedisplay(); break;
    case '6': scene.cam.girar( 0.1);       glutPostRedisplay(); break;
    case 'Z':
    case 'z': scene.cam.pan(false, true);  glutPostRedisplay(); break;
    case 'X':
    case 'x': scene.cam.pan(false, false); glutPostRedisplay(); break;
    case 'A':
    case 'a': scene.cam.pan(true,  false); glutPostRedisplay(); break;
    case 'D':
    case 'd': scene.cam.pan(true,  true);  glutPostRedisplay(); break;
    case 'F':
    case 'f': scene.cam.rx += SUMRX_INC; break;
    case '+': sumrx += 0.002; break;
    case '-': sumrx -= 0.002; break;
    case '/': if(scene.drawlevel > 0)                  scene.drawlevel--; glutPostRedisplay(); break;
    case '*': if(scene.drawlevel < scene.maxdrawlevel) scene.drawlevel++; glutPostRedisplay(); break;
    case '1': scene.drawmirror = !scene.drawmirror;             glutPostRedisplay(); break;
    case '2': scene.drawshadow = !scene.drawshadow;             glutPostRedisplay(); break;
    case '3': scene.drawmirrorshadow = !scene.drawmirrorshadow; glutPostRedisplay(); break;
    case 'Y':
    case 'y': scene.cam.interactiva = !scene.cam.interactiva;   glutPostRedisplay(); break;
    case 'H':
    case 'h': ShowHelp(); break;
    case '<': scene.RenderNFaces--; glutPostRedisplay(); break;
    case '>': scene.RenderNFaces++; glutPostRedisplay(); break;
    case ',': if(scene.cam.zoom > 0) scene.cam.zoom -= 0.01; else scene.cam.zoom = 0.01; scene.cam.SetupProjectionOpenGL(); glutPostRedisplay(); break;
    case '.': scene.cam.zoom += 0.01; scene.cam.SetupProjectionOpenGL(); glutPostRedisplay(); break;
    case '#': scene.cam.zoom = 1; scene.cam.SetupProjectionOpenGL(); glutPostRedisplay(); break;
    case '0': break;
    case 27:
    case 'Q':
    case 'q': cout << "\nEjecucin finalizada con exito\n"; exit(0);
    default:
      sumrx = (sumrx != 0) ? 0 : SUMRX_INC;
  }
}

void callbackMenu(int entry)
{
  ComandosInteractivos(entry);
}

void callbackKeyboard(unsigned char key, int x, int y )
{
  ComandosInteractivos(key);
}

void callbackMousePress(int button, int state, int x, int y)
{
  if(state == GLUT_DOWN)
  {
    if(button == GLUT_MIDDLE_BUTTON)
      doing_pan = 1;
  } else {
    if(button == GLUT_MIDDLE_BUTTON)
      doing_pan = 0;
  }
  last_x = x;
  last_y = y;
}

void callbackMouseDrag(int x, int y)
{
  if(doing_pan)
  {
    scene.cam.setVel(fabs(((float) (y - last_y)) / (SUAVIDAD_PAN))); scene.cam.pan(false, (y - last_y) >= 0);
    scene.cam.setVel(fabs(((float) (x - last_x)) / (SUAVIDAD_PAN))); scene.cam.pan(true,  (x - last_x) >= 0);
  } else {
    scene.cam.girar(((float) (x - last_x)) / (SUAVIDAD_GIRO), ((float) (y - last_y)) / (SUAVIDAD_GIRO));
  }
  last_x = x;
  last_y = y;
  glutPostRedisplay();
}

//////////////////////////////////////////////////////////////////////////////
// MENUS
//////////////////////////////////////////////////////////////////////////////

#define MENU_SEP          "-------------------------------------------------------"
void SetupMenus(void)
{
  glutCreateMenu(callbackMenu);
  
  glutAddMenuEntry("Opciones:",                                 0 );
  glutAddMenuEntry(MENU_SEP,                                    0 );
  glutAddMenuEntry("Activar/desactivar espejo",                '1');
  glutAddMenuEntry("Activar/desactivar sombras.",              '2');
  glutAddMenuEntry("Activar/desactivar sombras en el reflejo", '3');
  glutAddMenuEntry(MENU_SEP,                                    0 );
  glutAddMenuEntry("Cam. automatica/interactiva",              'y');
  glutAddMenuEntry(MENU_SEP,                                    0 );
  glutAddMenuEntry("Ayuda!!!!!",                               'h');
  glutAddMenuEntry(MENU_SEP,                                    0 );
  glutAddMenuEntry("Salir",                                    'q');
  
  glutAttachMenu(GLUT_RIGHT_BUTTON);
}
#undef MENU_SEP

//////////////////////////////////////////////////////////////////////////////
// AYUDA
//////////////////////////////////////////////////////////////////////////////

void ShowHelp(void)
{
  FILE *f;
  int c, c1;
  
  if((f = fopen("readme.txt", "r")) == NULL)
  {
    printf("No pude abrir el fichero de ayuda 'readme.txt'.\n");
    return;
  }
  
  printf("\n\n- AYUDA ---------------------------------------------------------------------\n");
  c1 = c = '\0';
  while(((c = fgetc(f)) != EOF) && (c != '*')) c1 = c;
  
  while((c = fgetc(f)) != EOF)
  {
    if((c1 == '\n') && (c == '*'))
      break;
    putchar(c);
    c1 = c;
  }
  
  printf("-----------------------------------------------------------------------------\n");
  
  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////
// MODELAJE Y/O CARGA DE LA ESCENA
//
// Hay dos posibilidades:
//   - Que la escena no este cacheada en disco
//         => Modelamos la escena y la grabamos a disco
//            (NOTA: Se generan archivos enormes (del orden de 20Mb)
//
//   - Que la escena este cacheada en disco
//         => Nos ahorramos los calculos y cargamos toda la info de disco
//
//////////////////////////////////////////////////////////////////////////////

void LoadModels(void)
{
    // Cargamos/Modelamos los objetos que usaremos
  bishop = LoadVRML("bishop.wrl", "bishop", Material3D());
  king   = LoadVRML("king.wrl",   "king",   Material3D());
  knight = LoadVRML("knight.wrl", "knight", Material3D());
  pawn   = LoadVRML("pawn.wrl",   "pawn",   Material3D());
  queen  = LoadVRML("queen.wrl",  "queen",  Material3D());
  rook   = LoadVRML("rook.wrl",   "rook",   Material3D());
  board  = ModelaTableroAjedrez("board",
                                ANCHURA_BORDE_TABLERO, ALTURA_BORDE_TABLERO, PIECE_SEPARATION, FACTOR_HUNDIMIENTO,
                                twhite, tblack, madeira);
  scene = Model3D("scene");
}

void AddPiece(int i, int j, Model3D *mod, Material3D *mat, int dl)
{
  scene.AddModel(*mod, Vector3D(((float) (j - 3)) * PIECE_SEPARATION -PIECE_SEPARATION*1.0/2.0,
                           -mod->BoundingBoxMin.Y(),
                           ((float) (i - 3)) * PIECE_SEPARATION -PIECE_SEPARATION*1.0/2.0), *mat, dl);
}

void ModelarEscena(void)
{
  int i, j, dl = 0;
  scene.reset();
    // Empezamos a construir la escena
  cout << "Generando la escena...\n";
  cout << "  - Generando el tablero.\n";
  scene.AddModel(board, Vector3D(0, 0, 0), dl++);     // Aadimos el tablero respetando los materiales
    // Renderizamos las piezas negras
  cout << "  - Generando las piezas negras.\n";
  for(i = 0; i < 8; i++)
    for(j = 0; j < 8; j++)
      if(Tablero[i*8 + j] > 0 && Tablero[i*8 + j] < 7)
        AddPiece(i, j, modelos[Tablero[i*8 + j] - 1], &pblack, dl++);
  
    // Le damos la vuelta a las piezas
  cout << "  - Invirtiendo las piezas.\n";
  pawn.FlipY(); knight.FlipY(); bishop.FlipY(); rook.FlipY(); queen.FlipY(); king.FlipY();
  
    // Renderizamos las blancas
  cout << "  - Generando las piezas blancas.\n";
  for(i = 0; i < 8; i++)
    for(j = 0; j < 8; j++)
      if(Tablero[i*8 + j] > 10 && Tablero[i*8 + j] < 17)
        AddPiece(i, j, modelos[Tablero[i*8 + j] - 11], &pwhite, dl++);
  
  cout << "  - Liberando memoria (borramos las piezas y tablero de memoria)\n";
  board.reset(); pawn.reset(); knight.reset(); bishop.reset(); rook.reset(); queen.reset(); king.reset();
  
  cout << "  - Calculando la caja y esfera contenedora de la escena...\n";
  scene.CalculateBoundings();
  
  cout << "  - Generando las sombras.\n";
  shadowvols = scene.BuildVolumetricShadows("shadows");
  
  cout << "  - Finalizando configuracion de la escena.\n";
  scene.shadows = &shadowvols;
  scene.drawmirror = true;
  scene.drawlevel = dl;
  scene.maxdrawlevel = dl;

  cout << "  - Informacion sobre la escena.\n";
  scene.print("    ");
  cout << "  - Salvando la escena a disco (CACHE).\n";
#ifdef ZIP_CACHE
  scene.saveZipped(SCENE_O3D, "scene");
#else
  scene.save(SCENE_O3D, "scene");
#endif
  cout << "  - Salvando las sombras a disco (CACHE).\n";
#ifdef ZIP_CACHE
  shadowvols.saveZipped(SHADOWVOLS_O3D, "shadows");
#else
  shadowvols.save(SHADOWVOLS_O3D, "shadows");
#endif
  cout << "Fin!\n";
}

//////////////////////////////////////////////////////////////////////////////
// PROGRAMA PRINCIPAL
//////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
  bool loading_failed = false;
    // Inicializamos OpenGL
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
  glutInitWindowSize(500, 500);
  glutInit(&argc, argv);
  glutCreateWindow("glChess3f(Gerardo, Garca, Pea)");
  SetupMenus();
  glutDisplayFunc(callbackRender);
  glutReshapeFunc(callbackResize);
  glutKeyboardFunc(callbackKeyboard);
  glutMouseFunc(callbackMousePress);
  glutMotionFunc(callbackMouseDrag);
  glutIdleFunc(callbackMove);
  glutTimerFunc(TIMER_MSECS, timerCallback, 0);

  try {
    cout << "Intentando cargar la escena ['" SCENE_O3D "']... ";
#ifdef ZIP_CACHE
    scene.loadZipped(SCENE_O3D);
#else
    scene.load(SCENE_O3D);
#endif
    cout << "OK!\n";
    cout << "Intentando cargar las sombras ['" SHADOWVOLS_O3D "']... ";
#ifdef ZIP_CACHE
    shadowvols.loadZipped(SHADOWVOLS_O3D);
#else
    shadowvols.load(SHADOWVOLS_O3D);
#endif
    cout << "OK!\n";
    scene.shadows = &shadowvols;
  } catch(O3Derror e) {
    if(e == O3Dfilenotexists)
    {
      scene.reset();
      shadowvols.reset();
      cout << "ERROR. No hay informacin cacheada.\n"; 
      loading_failed = true;
    } else {
      cout << "ERROR GRAVE. Interrumpimos ejecucin\n";
      exit(1);
    }
  }
  if(loading_failed)
  {
    cout << "Procedemos al calculo de la escena y sus sombras.\n";
      // Cargamos los modelos
    cout << "==== LOADING MODELS ========================================================\n";
    LoadModels();
      // Modelamos la escena
    cout << "==== MODELLING SCENE  ======================================================\n";
    ModelarEscena(); 
  }
    // Configuramos la camara
  scene.cam.setZnear(1);  // Necesitamos ponerlo cerca para la navegacin interactiva
  scene.cam.setZfar(scene.BoundingSphRad + scene.cam.dist*2.0);
           // Ponemos el maximo entre el doble de lo necesario y la suma de lo necesario teniendo en cuenta
           // que la navegacion interactiva interesa que te puedas alejar
    // Inicializamos el OpenGL
  scene.InitOpenGL();

  glutMainLoop(); /* MAI retorna...*/    
}
