#include<GL/glut.h>
#include<iostream>
#include"material3d.h"

Material3D::Material3D(void)
{
  onlyColor = false;
  twoSided = false;
  shiny = 1.0;
  initialized = false;
  mirror  = 0;
}

Material3D::Material3D(Color3D flc)
{
  onlyColor = true;
  flatColor = flc;
  alpha = flc.A();
  initialized = true;
  mirror  = 0;
}

Material3D::Material3D(Color3D amb, Color3D diff, Color3D spec, bool twoS, float sh, float al)
{
  onlyColor = false;
  ambient  = amb;
  diffuse  = diff;
  specular = spec;
  twoSided = false;
  shiny    = sh;
  alpha    = al;
  ambient.A(al);
  diffuse.A(al);
  specular.A(al);
  initialized = true;
  mirror  = 0;
}

Material3D::Material3D(Color3D amb, Color3D diff, Color3D spec, bool twoS, float sh, float al, int mirr)
{
  onlyColor = false;
  ambient  = amb;
  diffuse  = diff;
  specular = spec;
  twoSided = false;
  shiny    = sh;
  alpha    = al;
  ambient.A(al);
  diffuse.A(al);
  specular.A(al);
  mirror = mirr;
  initialized = true;
}

Material3D::Material3D(const Material3D &m)
{
  onlyColor = m.onlyColor;
  ambient = m.ambient;
  diffuse = m.diffuse;
  specular = m.specular;
  twoSided = m.twoSided;
  shiny = m.shiny;
  alpha = m.alpha;
  mirror = m.mirror;
  initialized = m.initialized;
}

void Material3D::MakeMirror(int mirr)
{
  if(!initialized)
  {
    initialized = true;
    twoSided = onlyColor = false;
    ambient = diffuse = specular = Color3D(0,0,0,0);
    shiny = alpha = 0;
  }
  mirror = mirr;
}

void Material3D::toOpenGL(void)
{
  GLenum face;
  
  if(onlyColor)
  {
    glColor3fv(flatColor.color);
  } else {
    face = (twoSided) ? GL_FRONT_AND_BACK : GL_FRONT;
      // Ponemos el ambient del material
    glMaterialfv(face, GL_AMBIENT,  ambient.color);
      // Ponemos el Diffuse
    glMaterialfv(face, GL_DIFFUSE,  diffuse.color);
      // Ponemos el Specular
    glMaterialfv(face, GL_SPECULAR, specular.color);
      // El parametro de Shiny
    glMaterialf(face,  GL_SHININESS, shiny);
  }
}

bool Material3D::operator == (Material3D &m)
{
  return (initialized == m.initialized && onlyColor == m.onlyColor &&
          mirror      == m.mirror      && flatColor == m.flatColor &&
          ambient     == m.ambient     && diffuse   == m.diffuse   &&
          specular    == m.specular    && twoSided  == m.twoSided  &&
          shiny       == m.shiny       && alpha     == m.alpha     &&
          mirror      == m.mirror);
}

void Material3D::print(char *sep = "")
{
  if(initialized)
  {
      if(mirror > 0)
      {
        cout << sep << "Material del espejo " << mirror << '\n';
      }
      if(onlyColor)
      {
        cout << sep << "Color: "; flatColor.println();
      } else {
        cout << sep << "Ambient:  "; ambient.println();
        cout << sep << "Difusse:  "; diffuse.println();
        cout << sep << "Specular: "; specular.println();
        cout << sep << "Shiny: " << shiny << "; Alpha: " << alpha << "; TwoSided: " << twoSided << '\n';
      }
  } else {
    cout << sep << "Material sin inicializar!\n";
  }
}

void Material3D::save(FILE *f)
{
  fwrite(&initialized, sizeof(initialized), 1, f);
  fwrite(&onlyColor, sizeof(onlyColor), 1, f);
  fwrite(&mirror, sizeof(mirror), 1, f);
  flatColor.save(f);
  ambient.save(f);
  diffuse.save(f);
  specular.save(f);
  fwrite(&twoSided, sizeof(twoSided), 1, f);
  fwrite(&shiny, sizeof(shiny), 1, f);
  fwrite(&alpha, sizeof(alpha), 1, f);
}

void Material3D::load(FILE *f)
{
  fread(&initialized, sizeof(initialized), 1, f);
  fread(&onlyColor, sizeof(onlyColor), 1, f);
  fread(&mirror, sizeof(mirror), 1, f);
  flatColor.load(f);
  ambient.load(f);
  diffuse.load(f);
  specular.load(f);
  fread(&twoSided, sizeof(twoSided), 1, f);
  fread(&shiny, sizeof(shiny), 1, f);
  fread(&alpha, sizeof(alpha), 1, f);
}
