#include<GL/glut.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/wait.h>
#include<fcntl.h>
#include<iostream>
#include<stdlib.h>
#include<stdio.h>
#include<unistd.h>
#include"model3d.h"
#include"util.h"

#define BUFFER_INC	3000
#define RENDER_N_FACES  10

Model3D::Model3D()
{
  nvertex = nfaces = nmat = 0; name = "";
  sbvertexl = sbfacel = sbmatl = 0;
  drawlevel = 0;
  shadows = NULL;
  drawmirrorshadow = drawmirror = drawshadow = false;
  RenderNFaces = RENDER_N_FACES;
}

Model3D::Model3D(string nm)
{
  nvertex = nfaces = nmat = 0; name = nm;
  sbvertexl = sbfacel = sbmatl = 0;
  drawlevel = 0;
  shadows = NULL;
  drawmirrorshadow = drawmirror = drawshadow = false;
  RenderNFaces = RENDER_N_FACES;
}

void Model3D::reset(void)
{
  int i;
  if(nvertex > 0)
  {
    for(i = 0; i < nvertex; i++)
      delete vertexlist[i];
    free(vertexlist);
  }
  
  if(nfaces > 0)
  {
    for(i = 0; i < nfaces; i++)
      delete facelist[i];
    free(facelist);
  }
  if(nmat > 0)
  {
    for(i = 0; i < nmat; i++)
      delete matlist[i];
    free(matlist);
  }
  sbvertexl = sbfacel = sbmatl = 0;
}

Material3D *Model3D::FindMaterial(Material3D &m)
{
  int i;
  for(i = 0; i < nmat; i++)
    if(m == *matlist[i])
      return matlist[i];

  return NULL;
}

Material3D *Model3D::FindOrAddMaterial(Material3D &m) throw(O3Derror)
{
  Material3D *pfind;
  pfind = FindMaterial(m);
  if(pfind != NULL)
    return pfind;
    // Reservamos memoria
  if(nmat == 0)
  {
    sbmatl = BUFFER_INC;
    if((matlist = (Material3D **) malloc(sizeof(Material3D *) * BUFFER_INC)) == NULL)
    {
      cout << "error:Model3D::FindOrAddMaterial():No hay memoria para la material list.\n";
      throw(O3Dnomem);
    }
  } else {
    if(nmat + 1 >= sbmatl)
    {
      sbmatl += BUFFER_INC;
      if((matlist = (Material3D **) realloc(matlist, sizeof(Material3D *) * sbmatl)) == NULL)
      {
        cout << "error:Model3D::FindOrAddMaterial():No hay memoria para la material list.\n";
        throw(O3Dnomem);
      }
    }
  }
    // Creamos el material
  matlist[nmat] = new Material3D(m);
  nmat++;
  
  return matlist[nmat - 1];
}

Vector3D *Model3D::FindVertex(Vector3D &p)
{
  int i;
  
  for(i = 0; i < nvertex; i++)
    if(p == *vertexlist[i])
      return vertexlist[i];
  
  return NULL;
}

Vector3D *Model3D::FindOrAddVertex(Vector3D &p) throw(O3Derror)
{
  Vector3D *pfind;
  
  pfind = FindVertex(p);
  if(pfind != NULL)
    return pfind;

    // Reservamos memoria  
  if(nvertex == 0)
  {
    sbvertexl = BUFFER_INC;
    if((vertexlist = (Vector3D **) malloc(sizeof(Vector3D *) * BUFFER_INC)) == NULL)
    {
      cout << "error:Model3D::FindOrAddVertex():No hay memoria para el vertex list.\n";
      throw(O3Dnomem);
    }
  } else {
    if(nvertex + 1 >= sbvertexl)
    {
      sbvertexl += BUFFER_INC;
      if((vertexlist = (Vector3D **) realloc(vertexlist, sizeof(Vector3D *) * sbvertexl)) == NULL)
      {
        cout << "error:Model3D::FindOrAddVertex():No hay memoria para el vertex list.\n";
        throw(O3Dnomem);
      }
    }
  }
  
    // Creamos el vertice
  vertexlist[nvertex] = new Vector3D(p);
  nvertex++;
  
  return vertexlist[nvertex - 1];
}

Face3D *Model3D::AddFace(Vector3D *ip0, Vector3D *ip1, Vector3D *ip2, Vector3D &n0, Vector3D &n1, Vector3D &n2, Material3D &m, int dl) throw(O3Derror)
{
  Material3D *mref;
    // Buscamos una referencia al material que vamos a usar
  mref = FindOrAddMaterial(m);
  
    // Reservamos memoria para la cara
  if(nfaces == 0)
  {
    sbfacel = BUFFER_INC;
    if((facelist = (Face3D **) malloc(sizeof(Face3D *) * BUFFER_INC)) == NULL)
    {
      cout << "error:Model3D::AddFace():No hay memoria para el face list.\n";
      throw(O3Dnomem);
    }
  } else {
    if(nfaces + 1 >= sbfacel)
    {
      sbfacel += BUFFER_INC;
      if((facelist = (Face3D **) realloc(facelist, sizeof(Face3D *) * sbfacel)) == NULL)
      {
        cout << "error:Model3D::AddFace():No hay memoria para el face list.\n";
        throw(O3Dnomem);
      }
    }
  }
  
    // Aadimos la cara a la lista de caras
  facelist[nfaces] = new Face3D(ip0, ip1, ip2, n0, n1, n2, mref, dl);
  nfaces++;
  
  return facelist[nfaces - 1];
}

Face3D *Model3D::ForceAddFace(Vector3D p0, Vector3D p1, Vector3D p2, Vector3D n0, Vector3D n1, Vector3D n2, Material3D &m, int dl) throw(O3Derror)
{
  Vector3D *ip0, *ip1, *ip2;
  
  ip0 = FindOrAddVertex(p0);
  ip1 = FindOrAddVertex(p1);
  ip2 = FindOrAddVertex(p2);

  return AddFace(ip0, ip1, ip2, n0, n1, n2, m, dl);
}

Face3D *Model3D::AddFace(Vector3D p0, Vector3D p1, Vector3D p2, Vector3D n0, Vector3D n1, Vector3D n2, Material3D &m, int dl) throw(O3Derror)
{
  Vector3D *ip0, *ip1, *ip2;
  int i;
  
  ip0 = FindOrAddVertex(p0);
  ip1 = FindOrAddVertex(p1);
  ip2 = FindOrAddVertex(p2);
  
  for(i = 0; i < nfaces; i++)
  {
    if(facelist[i]->P0 == ip0 && facelist[i]->P1 == ip1 && facelist[i]->P2 == ip2)
      throw(O3Dfaceexists);
  }
  
  return AddFace(ip0, ip1, ip2, n0, n1, n2, m, dl);
}

Face3D *Model3D::AddFace(Vector3D p0, Vector3D p1, Vector3D p2, int dl) throw(O3Derror)
{
  Material3D m; 
  Vector3D n;
  
  return AddFace(p0, p1, p2, n, n, n, m, dl);
}

Face3D *Model3D::ForceAddFace(Vector3D p0, Vector3D p1, Vector3D p2, int dl) throw(O3Derror)
{
  Material3D m;
  Vector3D n;
  
  return ForceAddFace(p0, p1, p2, n, n, n, m, dl);
}

Face3D *Model3D::AddFace(Vector3D p0, Vector3D p1, Vector3D p2, Vector3D n, Material3D &m, int dl) throw(O3Derror)
{
  return AddFace(p0, p1, p2, n, n, n, m, dl);
}

Face3D *Model3D::ForceAddFace(Vector3D p0, Vector3D p1, Vector3D p2, Vector3D n, Material3D &m, int dl) throw(O3Derror)
{
  return ForceAddFace(p0, p1, p2, n, n, n, m, dl);
}

Face3D *Model3D::AddFace(Vector3D p0, Vector3D p1, Vector3D p2, Material3D &m, int dl) throw(O3Derror)
{
  Vector3D n = Vector3D();
  return AddFace(p0, p1, p2, n, n, n, m, dl);
}

Face3D *Model3D::ForceAddFace(Vector3D p0, Vector3D p1, Vector3D p2, Material3D &m, int dl) throw(O3Derror)
{
  Vector3D n = Vector3D();
  return ForceAddFace(p0, p1, p2, n, n, n, m, dl);
}

void Model3D::printdetails(void)
{
  int i;
  
  cout << "Objeto: " << name << "\n";
  cout << "Tiene " << nvertex << " vertices:\n\n";
  for(i = 0; i < nvertex; i++)
  {
    cout << "\t- Vertice " << i << ": "; vertexlist[i]->println();
  }
  
  cout << "\nCon " << nfaces << " caras:\n\n";
  for(i = 0; i < nfaces; i++)
  {
    cout << "\t- Cara " << i << ": "; facelist[i]->print();
  }
  
  cout << "\nCon " << nmat << " materiales:\n\n";
  for(i = 0; i < nmat; i++)
  {
    cout << "\t- Material " << i << ":\n"; matlist[i]->print("\t\t");
  }
}

void Model3D::print(char *sep = "")
{
  int i, total_mirror;
  cout << sep << "Objeto: " << name << "\n";
  cout << sep << "  - Informacin sobre la escena:\n";
  cout << sep << "      + Vertices:   " << nvertex << '\n';
  cout << sep << "      + Caras:      " << nfaces  << '\n';
  cout << sep << "      + Materiales: " << nmat    << '\n';
  cout << sep << "  - Numero de triangulos que se dibujaran:\n";
  cout << sep << "      + La escena solo:                                             " << nfaces   << '\n';
  ;
  total_mirror = 0;
  for(i = 0; i < nfaces; i++)
    if(facelist[i]->material->mirror) total_mirror++;
  if(total_mirror)
    cout << sep << "      + La escena con espejo:                                       " << nfaces*2 + total_mirror*2 << '\n';
  else
    cout << sep << "      - << no hay espejos >>\n";
  if(shadows != NULL)
  {
    cout << sep << "      + Numero de triangulos en las sombras:                        " << shadows->nfaces << '\n';
    cout << sep << "      + La escena con sombras:                                      " << nfaces + nfaces * 6 * 3 << '\n';
    cout << sep << "      + La escena con sombras y espejo:                             " << nfaces*2 + total_mirror*2 + nfaces * 6 * 3     << '\n';
    cout << sep << "      + La escena con sombras y espejo con sombras en los reflejos: " << nfaces*2 + total_mirror*2 + nfaces * 6 * 3 * 2 << '\n';
  } else {
    cout << sep << "      - << La escena no tiene las sombras calculadas >>\n";
  }
}

void Model3D::MirrorToOpenGL(int mirror, int mask)
{
  int i;
  Material3D *last_mat = NULL;
  Vector3D planeeq;

  if(!mask && drawmirror)
  {
    if(drawmirror)
      cout << "Warning:Model3D::MirrorToOpenGL():Para dibujar espejos necesitamos un stencil buffer.\n";
    drawmirror = false;
  } else {
    if(drawmirror)
    {
        // Dibujamos en el stencil el area dibujable (el espejo)
      glClearStencil(0);
      glClear(GL_STENCIL_BUFFER_BIT);
      glEnable(GL_STENCIL_TEST);
      glDisable(GL_LIGHTING);
      glStencilFunc(GL_ALWAYS, mask, 0xFFFFFFFFL);
      glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
      glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
      glDepthMask(GL_FALSE);
      glCullFace(GL_BACK);
      glBegin(GL_TRIANGLES);
      for(i = 0; i < nfaces; i++)
      {
        if(facelist[i]->material->mirror == mirror && facelist[i]->drawlevel <= drawlevel)
        {
          glNormal3fv(facelist[i]->N0.vector); glVertex3fv(facelist[i]->P0->vector);
          glNormal3fv(facelist[i]->N1.vector); glVertex3fv(facelist[i]->P1->vector);
          glNormal3fv(facelist[i]->N2.vector); glVertex3fv(facelist[i]->P2->vector);
          planeeq = facelist[i]->planeeq;
        }
      }
      glEnd();
      glEnable(GL_LIGHTING);
      glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
      glDepthMask(GL_TRUE);
      glStencilFunc(GL_NOTEQUAL, 0, 0xFFFFFFFFL);
      glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
  
        // Pillamos un plano de corte cualquiera del tablero y lo establecemos como plano de clipping
        // NOTA: Esto es necesario para que el tablero no se redibuje mal
      glEnable(GL_CLIP_PLANE0);
      glClipPlane(GL_CLIP_PLANE0, planeeq.dvector);
        // Renderizamos como si nada pero boca abajo el reflejo en el espejo
      glPushMatrix();
      glCullFace(GL_FRONT);
      glScalef(1, -1, 1);
      OpaqueToOpenGL(false);
      if(drawmirrorshadow)
        ShadowToOpenGL(mask);
      glPopMatrix();
      glDisable(GL_CLIP_PLANE0);
      glDisable(GL_STENCIL_TEST);
    }
  }
  
    // Situamos bien la luz
  light.toOpenGL();
    // Dibujamos la cara del espejo
  glCullFace(GL_BACK);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glDepthMask(GL_TRUE);
    
  glBegin(GL_TRIANGLES);
  for(i = 0; i < nfaces; i++)
  {
    if(facelist[i]->material->initialized && last_mat != facelist[i]->material)
    {
      if(facelist[i]->material->alpha < 1.0)
        facelist[i]->material->toOpenGL();
      last_mat = facelist[i]->material;
    }
    if(last_mat->mirror == mirror && last_mat->alpha < 1.0 && facelist[i]->drawlevel <= drawlevel)
    {
      glNormal3fv(facelist[i]->N0.vector);
      glVertex3fv(facelist[i]->P0->vector);
      glNormal3fv(facelist[i]->N1.vector);
      glVertex3fv(facelist[i]->P1->vector);
      glNormal3fv(facelist[i]->N2.vector);
      glVertex3fv(facelist[i]->P2->vector);
    }
  }
  glEnd();

  glDisable(GL_BLEND);
}

void Model3D::OpaqueToOpenGL(bool force_alpha)
{
  int i;
  Material3D *last_mat = NULL;
  
    // Situamos bien la luz
  light.toOpenGL();
    // Dibujamos toda la escena
  glBegin(GL_TRIANGLES);
  for(i = 0; i < nfaces; i++)
  {
    if(facelist[i]->material->initialized && last_mat != facelist[i]->material)
    {
      if(facelist[i]->material->alpha >= 1.0)
        facelist[i]->material->toOpenGL();
      last_mat = facelist[i]->material;
    }
    if((!last_mat->mirror && facelist[i]->material->alpha >= 1.0 || force_alpha) && facelist[i]->drawlevel <= drawlevel)
    {
      if((last_mat->mirror || facelist[i]->material->alpha < 1.0) && force_alpha)
      {
        glEnd();
        glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
        glBegin(GL_TRIANGLES);
      }
      glNormal3fv(facelist[i]->N0.vector); glVertex3fv(facelist[i]->P0->vector);
      glNormal3fv(facelist[i]->N1.vector); glVertex3fv(facelist[i]->P1->vector);
      glNormal3fv(facelist[i]->N2.vector); glVertex3fv(facelist[i]->P2->vector);
      if((last_mat->mirror || facelist[i]->material->alpha < 1.0) && force_alpha)
      {
        glEnd();
        glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
        glBegin(GL_TRIANGLES);
      }
    }
  }
  glEnd();
}

void Model3D::ShadowToOpenGL(int mask)
{
  int i, j = 0, n_rest;
  
  if(shadows != NULL)
  {
      // Si el stencil es de menos de 4 bits apaga y vamonos
    if(mask < (1 << 4))
    {
      cout << "Warning:Model3D::ShadowToOpenGL():Para dibujar sombras se necesita un stencil de 4 o mas bits\n";
      drawmirrorshadow = drawshadow = false;
      return;
    }
      // Preparamos todo para las sombras
    glDisable(GL_LIGHTING);
    glEnable(GL_STENCIL_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(0, 0, 0, 0.30);

    for(i = 0; i < shadows->nfaces; i += 6*RenderNFaces)
    {
        // Tenemos que vigilar si estamos en la ultima iteracin,
        // y calcular bien cuantas caras nos quedan
      n_rest = MIN(shadows->nfaces - i, 6*RenderNFaces);
        // Inicializamos el stencil buffer para hacer virguerias
      glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
      glDepthMask(GL_FALSE);
        // Incrementamos el stencil solo a partir de la mascara
      glStencilFunc(GL_LEQUAL, mask, 0xFFFFFFFFL);
      glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
      glCullFace(GL_BACK);
      glBegin(GL_TRIANGLES);
      for(j = 0; j < n_rest; j++)
      {
        if(shadows->facelist[i+j]->drawlevel <= drawlevel)
        {
          glVertex3fv(shadows->facelist[i+j]->P0->vector);
          glVertex3fv(shadows->facelist[i+j]->P1->vector);
          glVertex3fv(shadows->facelist[i+j]->P2->vector);
        }
      }
      glEnd();
        // Decrementamos el stencil si estamos en la mascara
      glStencilFunc(GL_LESS, mask, 0xFFFFFFFFL);
      glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
      glCullFace(GL_FRONT);
      glBegin(GL_TRIANGLES);
      for(j = 0; j < n_rest; j++)
      {
        if(shadows->facelist[i+j]->drawlevel <= drawlevel)
        {
          glVertex3fv(shadows->facelist[i+j]->P0->vector);
          glVertex3fv(shadows->facelist[i+j]->P1->vector);
          glVertex3fv(shadows->facelist[i+j]->P2->vector);
        }
      }
      glEnd();
        // Dibujamos las sombras alla donde podamos y de paso
        // dejamos el stencil como lo encontramos al principio,
        // pero capando las zonas donde ya hemos hecho sombra
      glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
      glStencilOp(GL_KEEP, GL_KEEP, GL_ZERO);
      glDisable(GL_CULL_FACE);
      glBegin(GL_TRIANGLES);
      for(j = 0; j < n_rest; j++)
      {
        if(shadows->facelist[i+j]->drawlevel <= drawlevel)
        {
          glVertex3fv(shadows->facelist[i+j]->P0->vector);
          glVertex3fv(shadows->facelist[i+j]->P1->vector);
          glVertex3fv(shadows->facelist[i+j]->P2->vector);
        }
      }
      glEnd();
      glEnable(GL_CULL_FACE);
    }
    glDisable(GL_STENCIL_TEST);
    glDepthMask(GL_TRUE);
    glEnable(GL_LIGHTING);
    glDisable(GL_BLEND);
  }
}

void Model3D::toOpenGL(void)
{
  int mask, i;

    // Averiguamos cuantos bits tiene el stencil
  glGetIntegerv(GL_STENCIL_BITS, &i);
    // Calculamos la mascara que usaremos en el stencil
  if(i > 0)
    mask = 1 << (i - 1);
  else
    mask = 0; // No hay stencil buffer!!!!!!!!!
    // Nos aseguramos que RenderNFaces sea correcto (esto sirve para las sombras)
  while(RenderNFaces > 1 && (RenderNFaces*6 >= mask))
  {
    cout << "Ajustando RenderNFaces = " << RenderNFaces << '\n';
    RenderNFaces--;
  }
  if(RenderNFaces <= 0)
  {
    cout << "Ajustando RenderNFaces = " << RenderNFaces << '\n';
    RenderNFaces = 1;
  }
    
    // Comprobamos el drawlevel
  if(drawlevel > maxdrawlevel) drawlevel = maxdrawlevel;
  if(drawlevel < 0) drawlevel = 0;

    // Situamos la camara
  cam.SetupCameraOpenGL();

    // Dibujamos el espejo
  MirrorToOpenGL(1, mask);

    // Dibujamos las caras opacas, marcando con el stencil su ubicacion
    // NOTA: Las caras transparentes no las dibujamos, pero las marcamos
    //       en el stencil, dejando constancia de su existencia
  glCullFace(GL_BACK);
  glClearStencil(0);
  glClear(GL_STENCIL_BUFFER_BIT);
  glEnable(GL_STENCIL_TEST);
  glStencilFunc(GL_ALWAYS, mask, 0xFFFFFFFFL);
  glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
  OpaqueToOpenGL(true);
  glDisable(GL_STENCIL_TEST);
    // Dibujamos las sombras (si toca)
  if(drawshadow)
    ShadowToOpenGL(mask);
}

void Model3D::CalculateBoundings(void)
{
  int i;
  
  if(nvertex <= 0)
  {
    BoundingBoxMin = Vector3D(0,0,0);
    BoundingBoxMax = Vector3D(0,0,0);
  } else {
    BoundingBoxMin = *vertexlist[0];
    BoundingBoxMax = *vertexlist[0];
    for(i = 1; i < nvertex; i++)
    {
      BoundingBoxMin.BoundingMin(*vertexlist[i]);
      BoundingBoxMax.BoundingMax(*vertexlist[i]);
    }
  }
  BoundingSphCenter = (BoundingBoxMin + BoundingBoxMax) / 2.0;
  BoundingSphRad    = (BoundingBoxMax - BoundingSphCenter).modulus();
}


void Model3D::FlipY(void)
{
  int i;
  for(i = 0; i < nvertex; i++)
  {
    vertexlist[i]->vector[0] = vertexlist[i]->vector[0] * -1;
    vertexlist[i]->vector[2] = vertexlist[i]->vector[2] * -1;
  }
  
  for(i = 0; i < nfaces; i++)
  {
    facelist[i]->N0.vector[0] = facelist[i]->N0.vector[0] * -1;
    facelist[i]->N0.vector[2] = facelist[i]->N0.vector[2] * -1;
    facelist[i]->N1.vector[0] = facelist[i]->N1.vector[0] * -1;
    facelist[i]->N1.vector[2] = facelist[i]->N1.vector[2] * -1;
    facelist[i]->N2.vector[0] = facelist[i]->N2.vector[0] * -1;
    facelist[i]->N2.vector[2] = facelist[i]->N2.vector[2] * -1;
  }
}

void Model3D::AddModel(Model3D &md, int dl) throw (O3Derror)
{
  AddModel(md, Vector3D(0,0,0), Material3D(), dl);
}

void Model3D::AddModel(Model3D &md, Vector3D t, int dl) throw (O3Derror)
{
  AddModel(md, t, Material3D(), dl);
}

void Model3D::AddModel(Model3D &md, Vector3D t, Material3D m, int dl) throw (O3Derror)
{
  int i;
  Face3D *face;
  string msg = "    - Aadiendo el modelo [" + md.name + "] al modelo [" + name + "]:\t";

    // Aadimos todas las caras del modelo 'md' al modelo
  for(i = 0; i < md.nfaces; i++)
  {
    if((i & 0x0FF) == 1) PrintPercent(msg, i, md.nfaces);
    face = ForceAddFace(*(md.facelist[i]->P0) + t, *(md.facelist[i]->P1) + t, *(md.facelist[i]->P2) + t,
                        md.facelist[i]->N0, md.facelist[i]->N1, md.facelist[i]->N2,
                        (md.facelist[i]->material->initialized) ? *(md.facelist[i]->material) : ((m.initialized) ? m : *(md.facelist[i]->material)), dl);
    face->normal  = md.facelist[i]->normal;
    face->normal.refreshd();
    face->planeeq = md.facelist[i]->planeeq;
    face->planeeq.refreshd();
  }
  PrintPercent(msg, i, md.nfaces);
}

void Model3D::InitOpenGL(void)
{
    // Color de fondo
  glClearColor(0.0, 0.0, 0.0, 0.0);
    // Necesitamos el z-buffer
  glEnable(GL_DEPTH_TEST);
    // Queremos sombras suavizadas
  glShadeModel(GL_SMOOTH);
    // Ademas el culling esta a la orden del dia
  glEnable(GL_CULL_FACE);
    // Siempre que dibujemos un poligono, lo queremos rellenado
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    // Inicializamos las luces
  light.InitLight();
    // Inicializamos la camara
  cam.SetupProjectionOpenGL();
}

Vector3D Model3D::ClipVector(Vector3D base, Vector3D v)
{
  Vector3D r;
  r = base + v;
  
  if(r.X() > BoundingBoxMax.X()) { v = v * (BoundingBoxMax.X() - base.Y()) / (v.X()); r = base + v; }
  if(r.X() < BoundingBoxMin.X()) { v = v * (BoundingBoxMin.X() - base.Y()) / (v.X()); r = base + v; }
  if(r.Y() > BoundingBoxMax.Y()) { v = v * (BoundingBoxMax.Y() - base.Y()) / (v.Y()); r = base + v; }
  if(r.Y() < BoundingBoxMin.Y()) { v = v * (BoundingBoxMin.Y() - base.Y()) / (v.Y()); r = base + v; }
  if(r.Z() > BoundingBoxMax.Z()) { v = v * (BoundingBoxMax.Z() - base.Y()) / (v.Z()); r = base + v; }
  if(r.Z() < BoundingBoxMin.Z()) { v = v * (BoundingBoxMin.Z() - base.Y()) / (v.Z()); r = base + v; }
  
  return r;
}

Model3D Model3D::BuildVolumetricShadows(char *name) throw (O3Derror)
{
  Model3D sm = Model3D(name);
  Vector3D pu1, pu2;
  Material3D shadowm = Material3D();
  int i;
  
  string msg = "    - Construyendo el modelo de sombras[";
  msg = msg + name + "]:\t";

  for(i = 0; i < nfaces; i++)
  {
    if((i & 0x0FF) == 1) PrintPercent(msg, i, nfaces);
    pu1 = (*(facelist[i]->P0) - light.position); pu1.normalize(); pu1 = ClipVector(*(facelist[i]->P0), pu1 * (BoundingSphRad * 2));
    pu2 = (*(facelist[i]->P1) - light.position); pu2.normalize(); pu2 = ClipVector(*(facelist[i]->P1), pu2 * (BoundingSphRad * 2));
    sm.ForceAddFace(*(facelist[i]->P1), *(facelist[i]->P0), pu1, shadowm, facelist[i]->drawlevel);
    sm.ForceAddFace(*(facelist[i]->P1), pu1,                pu2, shadowm, facelist[i]->drawlevel);
    
    pu1 = (*(facelist[i]->P1) - light.position); pu1.normalize(); pu1 = ClipVector(*(facelist[i]->P1), pu1 * (BoundingSphRad * 2));
    pu2 = (*(facelist[i]->P2) - light.position); pu2.normalize(); pu2 = ClipVector(*(facelist[i]->P2), pu2 * (BoundingSphRad * 2));
    sm.ForceAddFace(*(facelist[i]->P2), *(facelist[i]->P1), pu1, shadowm, facelist[i]->drawlevel);
    sm.ForceAddFace(*(facelist[i]->P2), pu1,                pu2, shadowm, facelist[i]->drawlevel);

    pu1 = (*(facelist[i]->P2) - light.position); pu1.normalize(); pu1 = ClipVector(*(facelist[i]->P2), pu1 * (BoundingSphRad * 2));
    pu2 = (*(facelist[i]->P0) - light.position); pu2.normalize(); pu2 = ClipVector(*(facelist[i]->P0), pu2 * (BoundingSphRad * 2));
    sm.ForceAddFace(*(facelist[i]->P0), *(facelist[i]->P2), pu1, shadowm, facelist[i]->drawlevel);
    sm.ForceAddFace(*(facelist[i]->P0), pu1,                pu2, shadowm, facelist[i]->drawlevel);
  }
  PrintPercent(msg, i, nfaces);
  
  return sm;
}

void Model3D::save(FILE *f, char *name) throw (O3Derror)
{
  int i;
  
  fprintf(f, "O3D%s\n", name);
  fwrite(&nvertex, sizeof(nvertex), 1, f);
  fwrite(&nfaces, sizeof(nfaces), 1, f);
  fwrite(&nmat, sizeof(nmat), 1, f);
  fwrite(&drawshadow, sizeof(drawshadow), 1, f);
  fwrite(&drawmirrorshadow, sizeof(drawmirrorshadow), 1, f);
  fwrite(&drawmirror, sizeof(drawmirror), 1, f);
  fwrite(&drawlevel, sizeof(drawlevel), 1, f);
  fwrite(&maxdrawlevel, sizeof(maxdrawlevel), 1, f);
  fwrite(&RenderNFaces, sizeof(RenderNFaces), 1, f);
  
  for(i = 0; i < nvertex; i++)
    vertexlist[i]->save(f);
  
  for(i = 0; i < nmat; i++)
    matlist[i]->save(f);

  for(i = 0; i < nfaces; i++)
    facelist[i]->save(f, nvertex, vertexlist, nmat, matlist);
  
  light.save(f);
}

void Model3D::load(FILE *f, char *fname) throw (O3Derror)
{
  char temp[1000];
  int i;

  fread(temp, sizeof(char), 3, f);
  temp[3] = '\0';
  if(strcmp("O3D", temp))
  {
    cout << "error:Model3D::LoadModel():'" << fname << "' is malformed.\n";
    return;
  }
  fscanf(f, "%s", temp); cout << "Loading '" << temp << "'... ";
  fgetc(f);
  fread(&nvertex,          sizeof(nvertex),          1, f); sbvertexl = nvertex;
  fread(&nfaces,           sizeof(nfaces),           1, f); sbfacel   = nfaces;
  fread(&nmat,             sizeof(nmat),             1, f); sbmatl    = nmat;    
  fread(&drawshadow,       sizeof(drawshadow),       1, f);
  fread(&drawmirrorshadow, sizeof(drawmirrorshadow), 1, f);
  fread(&drawmirror,       sizeof(drawmirror),       1, f);
  fread(&drawlevel,        sizeof(drawlevel),        1, f);
  fread(&maxdrawlevel,     sizeof(maxdrawlevel),     1, f);
  fread(&RenderNFaces,     sizeof(RenderNFaces),     1, f);

  if((vertexlist = (Vector3D   **) malloc(sizeof(Vector3D   **) * nvertex)) == NULL) { cout << "error:Model3D::LoadModel():No hay memoria para el vertex list.\n"; throw(O3Dnomem); }
  if((matlist    = (Material3D **) malloc(sizeof(Material3D **) * nmat))    == NULL) { cout << "error:Model3D::LoadModel():No hay memoria para el material list.\n"; throw(O3Dnomem); }
  if((facelist   = (Face3D     **) malloc(sizeof(Face3D     **) * nfaces))  == NULL) { cout << "error:Model3D::LoadModel():No hay memoria para el face list.\n"; throw(O3Dnomem); }
  
  for(i = 0; i < nvertex; i++)
  {
    vertexlist[i] = new Vector3D(0,0,0);
    vertexlist[i]->load(f);
  }
  for(i = 0; i < nmat; i++)
  {
    matlist[i] = new Material3D();
    matlist[i]->load(f);
  }

  for(i = 0; i < nfaces; i++)
  {
    facelist[i] = new Face3D();
    facelist[i]->load(f, vertexlist, matlist);
  }
  
  light.load(f);
}

void Model3D::saveZipped(char *fname, char *name) throw (O3Derror)
{
  int raw_data[2];
  int zip_data;
  int status;
  FILE *f;
  
  pipe(raw_data);
  
  if((zip_data = open(fname, O_CREAT | O_TRUNC | O_WRONLY, 0666)) < 0)
  {
    cout << "error:Model3D::SaveModel():Can't open '" << fname << "' for binary writing.\n";
    throw O3Dsave;
  }
  
  switch(fork())
  {
    case -1: fprintf(stderr, "ERROR EN EL FORK.\n"); exit(1);
    case 0: // hijo
      close(0);
      dup(raw_data[0]);
      close(1);
      dup(zip_data);
      close(raw_data[1]);
      execlp("gzip", "gzip", "-f", (char *) 0);
      fprintf(stderr, "ERROR EN EXECLP.\n");
      exit(1);
    default:
      close(raw_data[0]);
      close(zip_data);
      if((f = fdopen(raw_data[1], "wb")) == NULL)
      {
        cout << "Error extrao!\n";
        exit(1);
      }
      save(f, name);
      fclose(f);
      wait(&status);
  }
}

void Model3D::loadZipped(char *fname) throw (O3Derror)
{
  int raw_data[2];
  int zip_data;
  int status;
  FILE *f;
  
  reset();
  
  
  if((zip_data = open(fname, O_RDONLY)) < 0)
  {
    cout << "error:Model3D::loadZipped():Can't open '" << fname << "' or not exists.\n";
    throw O3Dfilenotexists;
  }

  pipe(raw_data);
  
  switch(fork())
  {
    case -1: fprintf(stderr, "ERROR EN EL FORK.\n"); exit(1);
    case 0: // hijo
      close(raw_data[0]);
      close(0);
      dup(zip_data);
      close(1);
      dup(raw_data[1]);
      execlp("gunzip", "gunzip", (char *) 0);
      fprintf(stderr, "ERROR EN EXECLP.\n");
      exit(1);
    default:
      close(raw_data[1]);
      close(zip_data);
      if((f = fdopen(raw_data[0], "rb")) == NULL)
      {
        cout << "Error extrao!\n";
        exit(1);
      }
      load(f, fname);
      fclose(f);
      wait(&status);
  }
  
  CalculateBoundings();
  
  shadows = NULL;
}

void Model3D::load(char *fname) throw (O3Derror)
{
  FILE *f;
  
  reset();
  
  if((f = fopen(fname, "rb")) == NULL)
  {
    cout << "error:Model3D::LoadModel():Can't open '" << fname << "' or not exists.\n";
    throw O3Dfilenotexists;
  }
  
  load(f, fname);
    
  CalculateBoundings();
  
  shadows = NULL;
}

void Model3D::save(char *fname, char *name) throw (O3Derror)
{
  FILE *f;
  
  if((f = fopen(fname, "wb")) == NULL)
  {
    cout << "error:Model3D::SaveModel():Can't open '" << fname << "' for binary writing.\n";
    throw O3Dsave;
  }
  
  save(f, name);
  
  fclose(f);
}
