#ifndef __MODEL3D_H__
#define __MODEL3D_H__

#include<string>

#include"material3d.h"
#include"vector3d.h"
#include"face3d.h"
#include"light3d.h"
#include"camera3d.h"

enum O3Derror { O3Dnomem, O3Dfaceexists, O3Dsave, O3Dfilenotexists };

class Model3D {
  private:
    Vector3D   ClipVector(Vector3D, Vector3D);
    
  public:
    Vector3D   **vertexlist;      // Lista con todos los vertices del modelo
    Face3D     **facelist;        // Lista con todas las caras de modelo
    Material3D **matlist;         // Lista con todos los materiales del modelo
    int        nvertex;           // Numero de vertices del modelo
    int        nfaces;            // Numero de caras del modelo
    int        nmat;              // Numero de materiales del modelo
    int        sbvertexl;         // Tamao real de los buffers
    int        sbfacel;
    int        sbmatl;
    
    bool       drawshadow;
    bool       drawmirrorshadow;
    bool       drawmirror;
    int        drawlevel;
    int        maxdrawlevel;
    int        RenderNFaces;      // Numero de sombras que se tiran a la vez

    string     name;              // Nombre del modelo. Elemento puramente decorativo.
    
    Light3D    light;             // Iluminacion de la escena
    Camera3D   cam;               // La camara que usamos
    
    Model3D    *shadows;          // Modelo de sombras
    
    Vector3D   BoundingBoxMin;    // Minima esquina de la bounding box
    Vector3D   BoundingBoxMax;    // Maxima esquina de la bounding box
    Vector3D   BoundingSphCenter; // Centro de la escena
    float      BoundingSphRad;    // Radio de la esfera contenedero

      // Constructoras
    Model3D();
    Model3D(string nm);
    Model3D(string nm, Material3D &dm) throw(O3Derror);
    
      // Liberador de memoria. Es importante usarlo cuando se debe
    void reset(void);
    
      // Operaciones para construir el modelo
    Vector3D *FindVertex(Vector3D &p);
    Vector3D *FindOrAddVertex(Vector3D &p) throw(O3Derror);
    Face3D *AddFace(Vector3D *ip0, Vector3D *ip1, Vector3D *ip2, Vector3D &n0, Vector3D &n1, Vector3D &n2, Material3D &m, int dl) throw(O3Derror);
    Face3D *AddFace(Vector3D p0,   Vector3D p1,   Vector3D p2,   Vector3D n0,  Vector3D n1,  Vector3D n2,  Material3D &m, int dl) throw(O3Derror);
    Face3D *AddFace(Vector3D p0,   Vector3D p1,   Vector3D p2,   Vector3D n,   Material3D &m, int dl) throw(O3Derror);
    Face3D *AddFace(Vector3D p0,   Vector3D p1,   Vector3D p2,   Material3D &m, int dl) throw(O3Derror);
    Face3D *AddFace(Vector3D p0,   Vector3D p1,   Vector3D p2, int dl) throw(O3Derror);
    Face3D *ForceAddFace(Vector3D p0,   Vector3D p1,   Vector3D p2,   Vector3D n0,  Vector3D n1,  Vector3D n2,  Material3D &m, int dl) throw(O3Derror);
    Face3D *ForceAddFace(Vector3D p0,   Vector3D p1,   Vector3D p2,   Vector3D n,   Material3D &m, int dl) throw(O3Derror);
    Face3D *ForceAddFace(Vector3D p0,   Vector3D p1,   Vector3D p2,   Material3D &m, int dl) throw(O3Derror);
    Face3D *ForceAddFace(Vector3D p0,   Vector3D p1,   Vector3D p2, int dl) throw(O3Derror);
    Material3D *FindMaterial(Material3D &mat);
    Material3D *FindOrAddMaterial(Material3D &mat) throw(O3Derror);
    
      // Operaciones para concatenar un modelo
    void AddModel(Model3D &md, int dl) throw (O3Derror);
    void AddModel(Model3D &md, Vector3D t, int dl) throw (O3Derror);
    void AddModel(Model3D &md, Vector3D t, Material3D m, int dl) throw (O3Derror);
    
      // Calculos curiosos
    void CalculateBoundings(void);
    Model3D BuildVolumetricShadows(char *name) throw (O3Derror);
    
      // Transformaciones en el modelo
    void FlipY(void);
    
      // Para cargar y salvar modelos precalculados
    void save(char *fname, char *name) throw (O3Derror);
    void load(char *fname) throw (O3Derror);
    void saveZipped(char *fname, char *name) throw (O3Derror);
    void loadZipped(char *fname) throw (O3Derror);
    void save(FILE *f, char *name)  throw (O3Derror);
    void load(FILE *f, char *fname) throw (O3Derror);
    
      // Rutinas de renderizacion
    void MirrorToOpenGL(int i, int mask);
    void OpaqueToOpenGL(bool force_alpha);
    void ShadowToOpenGL(int mask);
    void toOpenGL(void);
    
      // Rutinas de inicializacion de OpenGL
    void InitOpenGL(void);
    
    void print(char *sep="");
    void printdetails(void);
};

#endif
